# A solver based on Algorithmic Lovász local lemma: https://en.wikipedia.org/wiki/Algorithmic_Lov%C3%A1sz_local_lemma
# and 
import random
from pylll.decision import lll_decision

def is_satisfied(clause, assignment):
    for literal in clause:
        if literal * assignment[abs(literal) - 1] > 0:
            return True
    return False


# http://tcs.nju.edu.cn/slides/random2015/LLL.pdf


def boundary(R, dependency):
    boundary = set()
    for idx in R:
        neighbors = dependency[idx]
        for neighbor in neighbors:
            if neighbor not in R:
                boundary.add(neighbor)
    return boundary

def vars_related(event_idx, S_idx, instance):
    vars_event = set([abs(x) for x in instance[event_idx]])
    vars_S = set([abs(x) for idx in S_idx for x in instance[idx]])
    if vars_event.intersection(vars_S):
        return True
    else:
        return False



def lll_solver(instance, n, sample=False):
    if not sample:
        if lll_decision(instance, n):
                assignment = [1 if random.random() > 0.5 else -1 for x in range(1, n + 1)]
                solved = False
                while not solved:
                    solved = True
                    for clause in instance:
                        if is_satisfied(clause, assignment):
                            pass
                        else:
                            solved = False
                            literal = random.choice(clause)
                            assignment[abs(literal) - 1] = -assignment[abs(literal) - 1]
                            break
                return assignment
        else:
            return None
    else:
        satisfiable, dependency = lll_decision(instance, n, need_dependency=True)
        if satisfiable:
            assignment = [1 if random.random() > 0.5 else -1 for x in range(1, n + 1)]
            solved = False
            while not solved:
                solved = True
                R = set()
                # init resampled set
                for idx, clause in enumerate(instance) :
                    if is_satisfied(clause, assignment):
                        pass
                    else:
                        solved = False
                        R.add(idx)
                N = set()
                # update resampled set
                diff = boundary(R, dependency) - N
                while diff:
                    for event_idx in diff:
                        #
                        if vars_related(event_idx, R, instance):
                            R.add(event_idx)
                        else:
                            N.add(event_idx)
                    diff = boundary(R, dependency) - N
                # resampled all variables in R
                vars = set([abs(x) for idx in R for x in instance[idx]])
                # print(f'R: {R}')
                for var in vars:
                    var_idx = var - 1
                    assignment[var_idx] = assignment[var_idx] if random.random() > 0.5 else - assignment[var_idx]
            return assignment
        else:
            return None


from functools import reduce
import operator


def lll_decision(instance, n, eval_iter=20, need_dependency=False):
    """n: the variable numbers"""
    num_clauses = len(instance)
    # initial the probability for each clause (event)
    event_prob = [2 ** (-len(clause)) for clause in instance]
    # dependency analysis
    occurence = [[] for i in range(n + 1)]  # j in occurence[i] means varible x_i is in j-th clauses
    dependency = [[] for i in range(num_clauses)]
    for i in range(num_clauses):
        clause_len = len(instance[i])
        for j in range(clause_len):
            occurence[abs(instance[i][j])].append(i)

    for i in range(1, n + 1):
        len_i = len(occurence[i])
        if len_i <= 1:
            continue
        for j in range(len_i):
            for l in range(j + 1, len_i):
                dependency[occurence[i][j]].append(occurence[i][l])
                dependency[occurence[i][l]].append(occurence[i][j])

    for i in range(num_clauses):
        dependency[i] = set(dependency[i])
    
    satisfiable = True
    # initial alphas
    alphas = [x for x in event_prob]
    # constraint propagation
    for _ in range(eval_iter):
        if max(alphas) > 1:
            satisfiable = False
            break
        update_alphas = [
            event_prob[i] / reduce(operator.mul, [1 - alphas[j] for j in dependency[i]]) if dependency[i] else event_prob[i] for i in range(num_clauses) 
        ]
        alphas = update_alphas

    if need_dependency:
        return satisfiable, dependency
    else:
        return satisfiable



import random
from random import sample

# A random satisfiable k-sat instance generator based on lovasz local lemma.
# Noticed that the constrian is different with the original version and based on Theorem 1.3 in paper
# "Disproof of the neighborhood conjecture with implications to SAT"


def lll_generator(n, k):
    if n < k:
        print("n must be greater or equal to k for k-sat problem")
        return []

    d = int(2**k / (2.71829 * k))
    instance = []
    vars_quota = [d] * n

    valid_vars = [i for i in range(n) if vars_quota[i] > 0]
    while len(valid_vars) >= k:
        clause = sample(valid_vars, k)
        for x in clause:
            vars_quota[x] -= 1
        valid_vars = [i for i in range(n) if vars_quota[i] > 0]
        clause = [x + 1 for x in clause]
        clause = [x if random.random() > 0.5 else -x for x in clause]

        instance.append(clause)

    return instance



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='satlll',
    version="0.1.0",
    description='A Python tool that implements Lovasz Local Lemma\'s application on The Boolean Satisfiability (SAT) problem, which includes an instance geneartor,an algorithm (decision procedure), which decide that if a SAT instance satisfies in local lemma regime, and a solver can find solution or sample a uniform random solution in the local lemma regime.',
    author="wwen",
    author_email="wenwh96@gmail.com",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: OS Independent',
    ],
    keywords='lovasz sat',
    url='https://github.com/opensat/PyLLL',
    packages=find_packages(exclude=['example', '.ipynb_checkpoints', '.vscode', 'backup', 'formulas', 'images', 'industrial_formulas', 'unsolve_formulas', 'phase_transition.ipynb']),
    python_requires='>=3.5',
    install_requires=[],
    include_package_data=True,
    zip_safe=False,
)
